-----------------------------------------------------------------------------------------
--
-- Piece.lua
                                       
-----------------------------------------------------------------------------------------

local utils = require "Utils"
local mydata = require "mydata"


local Piece = { GRAY=-2, STICKY=-1, FIXED=0, G1=1, G2=2, G3=3, G4=4, G5=5, G6=6, G7=7, Blank=25} 
Piece.gamePiece = { Piece.G1, Piece.G2, Piece.G3, Piece.G4, Piece.G5, Piece.G6, Piece.G7 }


local STATE = { WAIT=-1, READY=0, GETSET=1, SETTLING=2, TOREMOVE=3, REMOVED=4}


local Base = mydata.Base
local xS = mydata.xS*0.8
local _cellWidth = mydata.HorizontalDistance
local _cellHeight = mydata.VerticalDistance



local STATE = { WAIT=-1, READY=0, GETSET=1, SETTLING=2, TOREMOVE=3, REMOVED=4}  




function Piece.newPiece(board, gameMode, kind, new, st)

    local _numRow = 9--mydata.level.board.numRow

    local _piece = Sprite.new()
    _piece.name = "gamePiece"
        
    local _sprite = Bitmap.new(Base:getTextureRegion(7)) --["gamea5"] = 66
    _sprite:setAnchorPoint(0.5, 0.5)
    _piece.baseSprite = _sprite

    _piece:addChild(_sprite)


    _piece.kind = kind
    _piece.board = board
    _piece.movable = kind ~= Piece.FIXED     
    _piece.removeReport = true -- if false, gameMode will not be notified of the removal of this piece
    _piece.checked = false
    _piece.dirty = true
    _piece.col, _piece.row = 0, 0
    _piece.prevCol, _piece.prevRow = -1, -1
    local _state = STATE.WAIT
    local _updated, _sideway, _moving = false, false, false
    _piece._actions = {}
    local _moveTo = nil

    ----my addition
    _piece.pieceWidth = _sprite:getWidth()*xS
    _piece.pieceHeight = _sprite:getHeight()*xS
    _piece:setScale(xS, xS)
    _piece.new = new

    --if kind=="Blank" then

    --else
        if kind<=7 and kind>=1 then

            local f = mydata.colorTable.colorSelected[kind]
            _sprite:setColorTransform(f[1]/255,f[2]/255,f[3]/255)
            local _icon = Bitmap.new(Base:getTextureRegion(kind)) --["Shape1"] = 43, ["Shape2"] = 44, ["Shape3"] = 45, ["Shape4"] = 46, ["Shape5"] = 47, ["Shape6"] = 48,
            _icon:setAnchorPoint(0.5, 0.5)
            _piece:addChild(_icon)
            _piece.baseIcon = _icon       -------------2nd display layer
            
            if bgSat==1 then
                _icon:setColorTransform(0.9,0.9,0.9)
            end
            _icon:setScale(0.8*xS, 0.8*xS)
            _piece.baseIcon:setColorTransform(1,1,1)

            if kind>=1 and kind<=3 then

                local base = Bitmap.new(Base:getTextureRegion(7)) --["gamea5"] = 7
                base:setAnchorPoint(0.5, 0.5)
                _piece:addChild(base)
                base:setScale(xS, xS)
                local f = mydata.colorTable.colorSelected[kind]
                base:setColorTransform(f[1]/255,f[2]/255,f[3]/255)
                base:setAlpha(0)
                _piece.selectBase = base

            end

    end

    

    function _piece:basePiece()
        self.baseSprite:setTextureRegion(Base:getTextureRegion(8)) 
        self.baseSprite:setScale(1.1*self.baseSprite:getScaleX(), 1.2*self.baseSprite:getScaleY())
    end


    function _piece:ready()
        self.checked = false
        _sideway = false
        _moving = false
        _updated = false
        _state = STATE.READY
        _piece.prevCol = -1
        _piece.prevRow = -1
        _piece.beingRemoved = nil
        self.state = _state
    end
        
                
    -- non-movable will always return true        
    function _piece:isSettling()
        return (_state == STATE.SETTLING) or (not _piece.movable)
    end
       


    local function _doesMoved(nextState)
        _state = nextState
        _piece.state = _state
        _moving = false
        _updated = false
    end
    
    
    local function _updateAction()

        for i=1,#_piece._actions do
            _piece._actions[i]:update()
        end
                
    end
    
    
    -- force this piece to move to a give spot on board
    function _piece:scheduleMove(col, row)
        _moveTo = {col, row}
    end
    
    
    -- try to move to empty spot in the row below.  it can also move horizontally, but only after it has 
    -- moved down at least one row
    function _piece:update()
        if not self.movable then return end


        if _moveTo ~= nil then
            board:slidePiece(self, _moveTo[1], _moveTo[2], utils.curry1(_doesMoved, STATE.SETTLING))
            _updateAction()
            _moveTo = nil
            return

        end


        
        if _moving or _updated then 
            _updateAction()
            return 
        end
        
        local moved = false      
        local col, row = self.col, self.row

        local function checkEmpty(col, row)
            local ball = board:getBasePiece(col, row)
            if ball and ball.empty==true then
                return true
            end
            return false
        end


        local function ballBelow(col, row)

            local bBelow = checkEmpty(col, row)

            if bBelow then

                ballBelow(col,row+1)
                --if col==4 then print(row) end

                --return

            elseif (not bBelow) and board:getPiece(col, row)==nil then

                --if col==4 then print(row) end

                return true

            elseif (not bBelow) and board:getPiece(col, row)~=nil then

                return false

            end

            --return 

        end

        
        if _state == STATE.READY or _state == STATE.SETTLING then
            local fallable

            local function function_name(col,row )
            
                if checkEmpty(col,row) then

                    function_name(col,row+1)

                else fallable = board:canFallTo(self, col, row, nil, true)       -- down

               end

            end

            if col%2==0 and row<=_numRow-3 then

                if not checkEmpty(col,row+1) then                
                    fallable = board:canFallTo(self, col, row+1, nil, true)       -- down
                    
                elseif checkEmpty(col,row+1) then

                    function_name(col,row+1)

                    
            



                end
                
            elseif col%2~=0 and row<=_numRow-2 then

                if not checkEmpty(col,row+1) then                
                    fallable = board:canFallTo(self, col, row+1, nil, true)       -- down

                elseif checkEmpty(col,row+1) then

                    function_name(col,row+1)
            
                end
            end


            if fallable~=true then

                if col%2==0 and row<=_numRow-2 then

                    if not checkEmpty(col-1,row+1) then

                        fallable = board:canFallTo(self, col-1, row+1)  -- lower left
                    end

                elseif col%2~=0 and row<=_numRow-2 then

                    if not checkEmpty(col-1,row) then

                        fallable = board:canFallTo(self, col-1, row)  -- lower left
                    end
                end
                             
            end


            if fallable~=true then
                if col%2==0 and row<=_numRow-2 then

                    if not checkEmpty(col+1,row+1) then     
                        fallable = board:canFallTo(self, col+1, row+1)  -- lower right
                    end                    

                elseif col%2~=0 and row<=_numRow-2 then

                    if not checkEmpty(col+1,row) then     
                        fallable = board:canFallTo(self, col+1, row)  -- lower right
                    end
                end 
                           
            end
          
            _state = (not fallable) and STATE.SETTLING or STATE.GETSET
            self.state = _state
            --if self.col==1 then print("updating ".._state) end      

        end
        
        if _state == STATE.GETSET then

            local function function_name1(col, row)

                if col%2==0 and row>_numRow-2 then return false end
                if col%2~=0 and row>_numRow-1 then return false end

                if checkEmpty(col,row) then

                    function_name1(col, row+1)

                else 

                    --if self.col==1 then print("sliung up self.row "..self.row.." to col "..col.." and row "..row) end

                    moved = board:slidePiece(self, col, row, utils.curry1(_doesMoved, STATE.SETTLING))


                end

                
            end


            -- try down
            if col%2==0 and row<=_numRow-3 then

                if not checkEmpty(col,row+1) then

                    moved = board:slidePiece(self, col, row+1, utils.curry1(_doesMoved, STATE.SETTLING))

                elseif checkEmpty(col,row+1) then


                    function_name1(col, row+1, utils.curry1(_doesMoved, STATE.SETTLING))


                end
                

            elseif col%2~=0 and row<=_numRow-2 then

                if not checkEmpty(col,row+1) then

                    moved = board:slidePiece(self, col, row+1, utils.curry1(_doesMoved, STATE.SETTLING))

                elseif checkEmpty(col,row+1) then

                    function_name1(col, row+1)

                    
                end
            end
            if moved then
                _sideway = false
            end
            
            -- try lower left, then right if needed
            if (not moved) then


                if col%2==0 and row<=_numRow-2 then
                    if not checkEmpty(col-1,row+1) then                    
                        moved = board:slidePiece(self, col-1, row+1, utils.curry1(_doesMoved, STATE.SETTLING), "left")   ----left
                    end
                elseif col%2~=0 and row<=_numRow-2 then
                    if not checkEmpty(col-1,row) then 
                        --if self.col==5 and self.row==5 then print("SLIDING self.row "..self.row.." to col "..col.." and row "..row) end    
                        moved = board:slidePiece(self, col-1, row, utils.curry1(_doesMoved, STATE.SETTLING), "left")
                    end
                end

                

                if not moved then                    
                    if col%2==0 and row<=_numRow-2 then
                        if not checkEmpty(col+1,row+1) then
                            --if col+1==1 and row+1==2 then print("AAAAA") end                   
                            moved = board:slidePiece(self, col+1, row+1, utils.curry1(_doesMoved, STATE.SETTLING), "right")   ------right
                        end
                        
                    elseif col%2~=0 and row<=_numRow-2 then
                        if not checkEmpty(col+1,row) then
                            --if col+1==1 and row==2 then print("AAAAAbbbb") end       
                            moved = board:slidePiece(self, col+1, row, utils.curry1(_doesMoved, STATE.SETTLING), "right")
                        end
                    end
                end

                if moved then
                    _sideway = true
                end
            end    
            
            -- try left and right, but only if thie piece has moved diagonally before
            --[[if _sideway and not moved then      
                moved = board:slidePiece(self, col-1, row, utils.curry1(_doesMoved, STATE.SETTLING ))
                if not moved then 
                    moved = board:slidePiece(self, col+1, row, utils.curry1(_doesMoved, STATE.SETTLING ))
                end
            end]]
            if not moved then 
                _state = STATE.SETTLING 
                self.state = _state
            end
        end
        
        if moved then
            _moving = true
            _updated = true
            
        end
        
        _updateAction()
    end
    
    return _piece    
end


return Piece
